from collections import OrderedDict

from .base import ApiBase
import logging

logger = logging.getLogger(__name__)


class VendorCredits(ApiBase):
    SIMPLE_FIELDS = [
        'accountingBookDetailList',
        'applied',
        'applyList',
        'autoApply',
        'billingAddress',
        'createdDate',
        'currencyName',
        'customFieldList',
        'exchangeRate',
        'expenseList',
        'internalId',
        'itemList',
        'lastModifiedDate',
        'memo',
        'taxDetailsList',
        'taxDetailsOverride',
        'taxPointDate',
        'taxRegOverride',
        'total',
        'tranDate',
        'tranId',
        'transactionNumber',
        'unApplied',
        'userTaxTotal',
        'userTotal',
        'nullFieldList',
    ]

    RECORD_REF_FIELDS = [
        'account',
        'billAddressList',
        'class',
        'createdFrom',
        'currency',
        'customForm',
        'department',
        'entityTaxRegNum',
        'location',
        'nexus',
        'postingPeriod',
        'subsidiary',
        'subsidiaryTaxRegNum',
    ]

    def __init__(self, ns_client):
        ApiBase.__init__(self, ns_client=ns_client, type_name='VendorCredit')

    def post(self, data) -> OrderedDict:
        assert data['externalId'], 'missing external id'
        vendor_credit = self.ns_client.VendorCredit(**data)

        vendor_credit['entity'] = self.ns_client.RecordRef(**(data['entity']))

        self.build_simple_fields(self.SIMPLE_FIELDS, data, vendor_credit)

        self.build_record_ref_fields(self.RECORD_REF_FIELDS, data, vendor_credit)

        logger.debug('able to create vendor credit = %s', vendor_credit)

        res = self.ns_client.upsert(vendor_credit)
        return self._serialize(res)




error_reference = {
    "expense_report": {
        'category_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid category reference key \d+ for entity \d+", 
            'keys': ['expense_category', 'employee']
        },
        'account_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid account reference key \d+ for subsidiary \d+", 
            'keys': ['account', 'subsidiary']
        },
        'project_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid customer reference key \d+ for entity \d+", 
            'keys': ['customer', 'employee']
        },
        'location_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid location reference key \d+ for subsidiary \d+", 
            'keys': ['location', 'subsidiary']
        },
        'department_reference_error': {
            'regex':r"An error occured in a upsert request: Invalid department reference key \d+ for subsidiary \d+" , 
            'keys': ['department', 'subsidiary']
        },
        'currency_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid currency reference key \d+ for subsidiary \d+", 
            'keys': ['currency', 'subsidiary']
        }
    },
    "bills": {
        'bill_account_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid account reference key \d+ for subsidiary \d+",
            'keys': ['account', 'subsidiary']
        },
        'location_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid location reference key \d+ for subsidiary \d+", 
            'keys': ['location', 'subsidiary']
        },
        'department_reference_error': {
            'regex':r"An error occured in a upsert request: Invalid department reference key \d+ for subsidiary \d+" , 
            'keys': ['department', 'subsidiary']
        },
        'currency_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid currency reference key \d+ for subsidiary \d+", 
            'keys': ['currency', 'subsidiary']
        },
        'vendor_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid entity reference key \d+ for subsidiary \d+", 
            'keys': ['vendor', 'subsidiary']
        }
    },
    "journal_entry": {
        'location_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid location reference key \d+ for subsidiary \d+", 
            'keys': ['location', 'subsidiary']
        },
        'department_reference_error': {
            'regex':r"An error occured in a upsert request: Invalid department reference key \d+ for subsidiary \d+" , 
            'keys': ['department', 'subsidiary']
        },
        'account_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid account reference key \d+ for subsidiary \d+", 
            'keys': ['account', 'subsidiary']
        },
        'currency_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid currency reference key \d+ for subsidiary \d+", 
            'keys': ['currency', 'subsidiary']
        },
        'project_reference_error': {
            'regex': r"An error occured in a upsert request: Invalid customer reference key \d+ for entity \d+", 
            'keys': ['customer', 'employee']
        },
    }
}


list_of_dicts = [
        ['expense_category', 'employee'], ['account', 'subsidiary'],
        ['customer', 'employee'], ['location', 'subsidiary'],
        ['department', 'subsidiary'], ['currency', 'subsidiary'],
        ['vendor', 'subsdiary']
    ]



"""
    :class:`NetSuiteError`
    :class:`NetSuiteLoginError`
    :class:`NetSuiteRequestError`
    :class:`NetSuiteTypeError`
    :class:`NetSuiteRateLimitError`
"""

class NetSuiteError(Exception):
    """Exception raised for errors during login or other requests (like
    get, getAll) to NetSuite."""

    def __init__(self, message, code=None):
        """
        :param str message: Text describing the error
        :param str code: Netsuite code specifying the exact error
                Possible values are listed in
        """

        self.message = message
        self.code = code

    def __str__(self):
        if self.code is None:
            return self.message
        return 'code: {}, message: {}'.format(self.code, self.message)


class NetSuiteLoginError(NetSuiteError):
    """Exception raised for errors during login to NetSuite"""

    pass

class NetSuiteRequestError(NetSuiteError):
    """Exception raised for errors during requests like get, search, .."""

    pass

class NetSuiteTypeError(NetSuiteError):
    """Exception raised when requested an invalid netsuite type"""

    pass

class NetSuiteRateLimitError(NetSuiteError):
    """Exception raised for errors during rate limit requests like get, search, .."""

    pass



from .helpers import replace_numbers, convert_to_camelcase
from .errors import list_of_dicts

class ErrorParser():

    def __init__(self, get_instance):
        self.get_instance = get_instance

    def get_entity_values(self, error_dict):

        entity_keys = list(error_dict)
        object_1 = self.get_instance(convert_to_camelcase(entity_keys[0]), error_dict[entity_keys[0]])
        object_2 = self.get_instance(convert_to_camelcase(entity_keys[1]), error_dict[entity_keys[1]])

        if object_1 and object_2:
            if entity_keys[0] == 'customer' and entity_keys[1] == 'employee':
                return object_1['entityId'], object_2['entityId']

            if entity_keys[1] == 'employee':
                return object_1['name'], object_2['entityId']

            if entity_keys[0] == 'account':
                object_1 = object_1['acctName']
                return object_1, object_2['name']

            if entity_keys[0] == 'vendor':
                return object_1['entityId'], object_2['name']

            return object_1['name'], object_2['name']


    def export_error_parser(self, error_dict, message):

        parsed_message = message
        if list(error_dict) in list_of_dicts:
            object_1, object_2 = self.get_entity_values(error_dict)
            entity_keys = list(error_dict)
            parsed_message = replace_numbers(message, object_1, object_2, error_dict[entity_keys[0]], error_dict[entity_keys[1]])
        return parsed_message



import re
from .errors import error_reference


def replace_numbers(string , replacement1, replacement2, number1, number2):
    pattern = r"\b({0}|{1})\b".format(number1, number2)
    replaced_string = re.sub(pattern, lambda match: replacement1 if match.group() == str(number1) else replacement2, string)
    return replaced_string


def convert_to_camelcase(word):
    return ''.join(word.title().split('_'))


def export_error_matcher(string, export_type):
    for _, error_data in error_reference[export_type].items():
        if re.match(error_data['regex'], string):
            numbers = re.findall(r'\d+', string)
            return {key: int(number) for key, number in zip(error_data['keys'], numbers)}
    
    return {}



import logging

from netsuitesdk.internal.utils import PaginatedSearch

from .base import ApiBase
from typing import List
from collections import OrderedDict

logger = logging.getLogger(__name__)


class VendorBills(ApiBase):
    """
    VendorBills are not directly searchable - only via as transactions
    """

    def __init__(self, ns_client):
        ApiBase.__init__(self, ns_client=ns_client, type_name='vendorBill')

    def get_all_generator(self):
        record_type_search_field = self.ns_client.SearchStringField(searchValue='VendorBill', operator='contains')
        basic_search = self.ns_client.basic_search_factory('Transaction', recordType=record_type_search_field)
        paginated_search = PaginatedSearch(client=self.ns_client,
                                           type_name='Transaction',
                                           basic_search=basic_search,
                                           pageSize=20)
        return self._paginated_search_to_generator(paginated_search=paginated_search)

    def post(self, data) -> OrderedDict:
        assert data['externalId'], 'missing external id'
        vb = self.ns_client.VendorBill(externalId=data['externalId'])
        # If expesess are present, add them 
        if 'expenseList' in data and data['expenseList']:
            expense_list = []
            for eod in data['expenseList']:
                if 'customFieldList' in eod and eod['customFieldList']:
                    custom_fields = []
                    for field in eod['customFieldList']:
                        if field['type'] == 'String':
                            custom_fields.append(
                                self.ns_client.StringCustomFieldRef(
                                    scriptId=field['scriptId'] if 'scriptId' in field else None,
                                    internalId=field['internalId'] if 'internalId' in field else None,
                                    value=field['value']
                                )
                            )
                        elif field['type'] == 'Select':
                            custom_fields.append(
                                self.ns_client.SelectCustomFieldRef(
                                    scriptId=field['scriptId'] if 'scriptId' in field else None,
                                    internalId=field['internalId'] if 'internalId' in field else None,
                                    value=self.ns_client.ListOrRecordRef(
                                        internalId=field['value']
                                    )
                                )
                            )
                    eod['customFieldList'] = self.ns_client.CustomFieldList(custom_fields)
                vbe = self.ns_client.VendorBillExpense(**eod)
                expense_list.append(vbe)

            vb['expenseList'] = self.ns_client.VendorBillExpenseList(expense=expense_list)

        # If items are present, add them
        if 'itemList' in data and data['itemList']:
            item_list = []
            for eod in data['itemList']:
                if 'customFieldList' in eod and eod['customFieldList']:
                    custom_fields = []
                    for field in eod['customFieldList']:
                        if field['type'] == 'String':
                            custom_fields.append(
                                self.ns_client.StringCustomFieldRef(
                                    scriptId=field['scriptId'] if 'scriptId' in field else None,
                                    internalId=field['internalId'] if 'internalId' in field else None,
                                    value=field['value']
                                )
                            )
                        elif field['type'] == 'Select':
                            custom_fields.append(
                                self.ns_client.SelectCustomFieldRef(
                                    scriptId=field['scriptId'] if 'scriptId' in field else None,
                                    internalId=field['internalId'] if 'internalId' in field else None,
                                    value=self.ns_client.ListOrRecordRef(
                                        internalId=field['value']
                                    )
                                )
                            )
                    eod['customFieldList'] = self.ns_client.CustomFieldList(custom_fields)
                vbe = self.ns_client.VendorBillItem(**eod)
                item_list.append(vbe)

            vb['itemList']= self.ns_client.VendorBillItemList(item=item_list)

        if 'currency' in data:
            vb['currency'] = self.ns_client.RecordRef(**(data['currency']))

        if 'memo' in data:
            vb['memo'] = data['memo']

        if 'tranDate' in data:
            vb['tranDate'] = data['tranDate']

        if 'tranId' in data:
            vb['tranId'] = data['tranId']

        if 'class' in data:
            vb['class'] = self.ns_client.RecordRef(**(data['class']))

        if 'location' in data:
            vb['location'] = self.ns_client.RecordRef(**(data['location']))

        if 'department' in data:
            vb['department'] = self.ns_client.RecordRef(**(data['department']))

        if 'account' in data:
            vb['account'] = self.ns_client.RecordRef(**(data['account']))

        if 'customFieldList' in data:
            vb['customFieldList'] = data['customFieldList']

        if 'entity' in data:
            vb['entity'] = self.ns_client.RecordRef(**(data['entity']))

        if 'taxDetailsOverride' in data:
            vb['taxDetailsOverride'] = data['taxDetailsOverride']

        if 'taxDetailsList' in data and data['taxDetailsList']:
            tax_details_list = []
            if 'taxDetails' in data['taxDetailsList']:
                for tdl in data['taxDetailsList']['taxDetails']:
                    tax_details_list.append(self.ns_client.TaxDetails(**tdl))
                vb['taxDetailsList'] = self.ns_client.TaxDetailsList(taxDetails=tax_details_list)

        logger.debug('able to create vb = %s', vb)
        res = self.ns_client.upsert(vb, 'bills')
        return self._serialize(res)



from .base import ApiBase
from netsuitesdk.internal.utils import PaginatedSearch
import logging

logger = logging.getLogger(__name__)

class Items(ApiBase):

    def __init__(self, ns_client):
        ApiBase.__init__(self, ns_client=ns_client, type_name='Item')

    def get_all_generator(self, is_inactive=False):
        # Get Only Active Items using SearchBooleanField
        record_type_search_field = self.ns_client.SearchBooleanField(searchValue=is_inactive)
        basic_search = self.ns_client.basic_search_factory('Item', isInactive=record_type_search_field)

        paginated_search = PaginatedSearch(
            client=self.ns_client,
            type_name='Item',
            basic_search=basic_search,
            pageSize=20
        )

        return self._paginated_search_generator(paginated_search=paginated_search)
